
// Some stuff copied from the Broadcom examples, ofcoz!

#include <bcm_host.h>
#include <GLES/gl.h>
#include <EGL/egl.h>
#include <EGL/eglext.h>

#include "koelli.h"

static EGLDisplay display;
static EGLSurface surface;
static EGLContext context;
static GLuint texture;
static uint32_t width,height,buf_width,buf_height,flagga,depthi;

static GLfloat koord[]={-1,-1,0, 1,-1,0, 1,1,0, -1,1,0},
               tkoord[]={0,1, 1,1, 1,0, 0,0};

void koelli_open(int xres,int yres,int depth,int flags)
{
    static const EGLint attribute_list[] =
    {
        EGL_RED_SIZE, 8,
        EGL_GREEN_SIZE, 8,
        EGL_BLUE_SIZE, 8,
        EGL_ALPHA_SIZE, 8,
        EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
        EGL_NONE
    };

    EGLint num_config;
    EGLConfig config;

    static EGL_DISPMANX_WINDOW_T nativewindow;

    // Sheesh. Could it be a bit MORE complicated?
    DISPMANX_ELEMENT_HANDLE_T dispman_element;
    DISPMANX_DISPLAY_HANDLE_T dispman_display;
    DISPMANX_UPDATE_HANDLE_T dispman_update;
    VC_RECT_T dst_rect;
    VC_RECT_T src_rect;

    int hx,hy;

    buf_width=xres;
    buf_height=yres;
    flagga=flags;
    depthi=depth;

    bcm_host_init();

    display=eglGetDisplay(EGL_DEFAULT_DISPLAY);
    eglInitialize(display,NULL,NULL);
    eglChooseConfig(display,attribute_list,&config,1,&num_config);
    context=eglCreateContext(display,config,EGL_NO_CONTEXT,NULL);

    graphics_get_display_size(0,&width,&height);

    dst_rect.x=0;
    dst_rect.y=0;
    dst_rect.width=width;
    dst_rect.height=height;

    src_rect.x=0;
    src_rect.y=0;
    src_rect.width=width<<16;
    src_rect.height=height<<16;

    dispman_display=vc_dispmanx_display_open(0);
    dispman_update=vc_dispmanx_update_start(0);

    dispman_element=vc_dispmanx_element_add(dispman_update,
                    dispman_display, 0/*layer*/, &dst_rect, 0/*src*/,
                    &src_rect, DISPMANX_PROTECTION_NONE, 0 /*alpha*/,
                    0/*clamp*/, 0/*transform*/);

    nativewindow.element=dispman_element;
    nativewindow.width=width;
    nativewindow.height=height;
    vc_dispmanx_update_submit_sync(dispman_update);

    surface = eglCreateWindowSurface(display,config,&nativewindow,NULL);

    eglMakeCurrent(display,surface,surface,context);

    glClearColor(0,0,0,1);
    glClear(GL_COLOR_BUFFER_BIT+GL_DEPTH_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE,GL_ONE);

    glGenTextures(1,&texture);
    glBindTexture(GL_TEXTURE_2D,texture);

    if(depth==32)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, KOEL_MAXX, KOEL_MAXY, 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    else
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, KOEL_MAXX, KOEL_MAXY, 0,
                     GL_RGBA, GL_UNSIGNED_SHORT_5_6_5, NULL);

    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

    // Scale window and coords according to the buffer and resolution
    GLfloat x=(GLfloat)xres/(GLfloat)KOEL_MAXX,
            y=(GLfloat)yres/(GLfloat)KOEL_MAXY;
    tkoord[1]=y; tkoord[2]=x; tkoord[3]=y; tkoord[4]=x;

    hx=width/2,
    hy=height/2;

    // Video output! Need borders.
    if((width==720 && height==576) || (width==720 && height==480))
    {
        hx=hx*88/100;
        hy=hy*88/100;
    }
    else
    {
        hx=height*xres/yres/2; // Probably only 4:3 modes work for now
    }

    glViewport(width/2-hx,height/2-hy,hx*2,hy*2);
}

void koelli_close(void)
{
    glClear(GL_COLOR_BUFFER_BIT);
    eglSwapBuffers(display,surface);

    eglMakeCurrent(display,EGL_NO_SURFACE,EGL_NO_SURFACE,EGL_NO_CONTEXT);

    eglDestroySurface(display,surface);
    eglDestroyContext(display,context);
    eglTerminate(display);
}

void koelli_blit(void *buf)
{
    int i;
    unsigned u,*p;

    if(flagga&KOEL_SWAP)
    {
        p=buf;
        for(i=0;i<buf_width*buf_height;i++,p++)
        {
            u=*p;
            *p=(u&0xff00)+((u&0xff)<<16)+((u&0xff0000)>>16);
        }
    }

    glClear(GL_COLOR_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glColor4f(1,1,1,1);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,texture);

    if(depthi==32)
        glTexSubImage2D(GL_TEXTURE_2D,0, 0,0, buf_width,buf_height,
                        GL_RGBA,GL_UNSIGNED_BYTE,buf);
    else
        glTexSubImage2D(GL_TEXTURE_2D,0, 0,0, buf_width,buf_height,
                        GL_RGBA,GL_UNSIGNED_SHORT_5_6_5,buf);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3,GL_FLOAT,0,koord);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glTexCoordPointer(2,GL_FLOAT,0,tkoord);
    glDrawArrays(GL_TRIANGLE_FAN,0,4);

    glDisable(GL_TEXTURE_2D);

    eglSwapBuffers(display,surface);
}
