/* mp3.c
 *
 * Copyright 2001-2003 Vesa Halttunen (Vesuri/dA JoRMaS)
 *
 * This file is part of JRm-core.
 *
 * JRm-core is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JRm-core is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JRm-core; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <SDL.h>
#ifdef WINDOWS
#include <Windows.h>
#else
#include <sys/time.h>
#endif
#include "mpglib/mpg123.h"
#include "mpglib/mpglib.h"
#include "mp3.h"

static struct mpstr *mp=NULL;
static FILE **file=NULL;
static char **inbuf=NULL;
static int *inbufsize=NULL, *decoded=NULL, *ret=NULL;
static int starttimereset=0, numfiles=0;
static char **decodebuf=NULL;
static long played;
extern int quit;
#ifdef WINDOWS
extern DWORD starttc;
#else
extern struct timeval starttv;
extern struct timezone tz;
#endif

int openmp3(char *filename, int outbufsize, int skipms) {
  /* Open MP3 file */
  FILE *f=fopen(filename, "rb");

  if(f) {
    int decode=skipms*1764/10+outbufsize;

    numfiles++;
    mp=(struct mpstr *)realloc(mp, numfiles*sizeof(struct mpstr));
    file=(FILE **)realloc(file, numfiles*sizeof(FILE *));
    inbuf=(char **)realloc(inbuf, numfiles*sizeof(char *));
    inbufsize=(int *)realloc(inbufsize, numfiles*sizeof(int));
    decoded=(int *)realloc(decoded, numfiles*sizeof(int));
    ret=(int *)realloc(ret, numfiles*sizeof(int));
    decodebuf=(char **)realloc(decodebuf, numfiles*sizeof(char *));
    
    file[numfiles-1]=f;
    inbuf[numfiles-1]=NULL;
    ret[numfiles-1]=MP3_NEED_MORE;
    decodebuf[numfiles-1]=calloc(8*outbufsize, sizeof(char));
    played=0;

    /* Initialize MP3 recoder */
    InitMP3(&mp[numfiles-1]);

    /* Read the input file */
    fseek(file[numfiles-1], 0, SEEK_END);
    inbufsize[numfiles-1]=ftell(file[numfiles-1]);
    inbuf[numfiles-1]=(char *)calloc(inbufsize[numfiles-1], 1);
    fseek(file[numfiles-1], 0, SEEK_SET);
    fread(inbuf[numfiles-1], 1, inbufsize[numfiles-1], file[numfiles-1]);

    decoded[numfiles-1]=outbufsize;
    while(decode>0) {
      memmove(decodebuf[numfiles-1]+(decoded[numfiles-1]-outbufsize),
	      decodebuf[numfiles-1], outbufsize);
      decoded[numfiles-1]-=outbufsize;

      while(decoded[numfiles-1]<outbufsize) {
	int add=0;
	if(ret[numfiles-1]!=MP3_OK)
	  ret[numfiles-1]=decodeMP3(&mp[numfiles-1], inbuf[numfiles-1],
				    inbufsize[numfiles-1],
				    decodebuf[numfiles-1]+decoded[numfiles-1],
				    32768, &add);
	else
	  ret[numfiles-1]=decodeMP3(&mp[numfiles-1], NULL, 0,
				    decodebuf[numfiles-1]+decoded[numfiles-1],
				    32768, &add);
	
	if(ret[numfiles-1]==MP3_OK)
	  decoded[numfiles-1]+=add;
	decode-=add;
      }
    }

    return 0;
  } else
    return -1;
}

/* Copies decoded data to the SDL buffer */
void playmp3(void *userdata, char *outbuf, int outbufsize) {
  struct playmp3data *data=(struct playmp3data *)userdata;
  long curtime;
  long *time=data->time;
  int *volume=data->volume, i;
  short *ob=(short *)outbuf, t;

  if(starttimereset<1) {
    /* Get the accurate time when exactly two buffers are played - the
     * first one is empty */
#ifdef WINDOWS
    starttc=GetTickCount()+outbufsize*10/441/4;
#else
    gettimeofday(&starttv, &tz);
    
    starttv.tv_usec+=outbufsize*10000/441/4;
    if(starttv.tv_usec>999999) {
      starttv.tv_sec++;
      starttv.tv_usec-=1000000;
    }
#endif
    starttimereset++;
  }

  curtime=played*10/441/4;

  for(i=0; i<numfiles; i++) {
    int j;

    if(curtime+(outbufsize*10000/441/4/1000)>=time[i] && volume[i]) {
      short *ib=(short *)decodebuf[i];
      int tocopy=outbufsize;
      int s;

      if(curtime-time[i]<0)
	tocopy=((time[i]-curtime)*4*44100/1000)&0xfffffffc;

      /*
      fprintf(stderr, "clock: %ld begin: %ld difference: %ld copying: %d\n",
	      curtime, time[i], curtime-time[i], tocopy);
      */

      for(j=0; j<tocopy/2; j++) {
	s=ob[j+outbufsize/2-tocopy/2]+volume[i]*((long)ib[j])/256;
	if(s>SHRT_MAX)
	  s=SHRT_MAX;
	else if(s<SHRT_MIN)
	  s=SHRT_MIN;
	   
	ob[j+outbufsize/2-tocopy/2]=s;
      }

      memmove(decodebuf[i], decodebuf[i]+tocopy, decoded[i]-tocopy);
      decoded[i]-=tocopy;
      
      while(volume[i] && decoded[i]<outbufsize) {
	int add=0;
	
	if(ret[i]!=MP3_OK)
	  /*
	  ret[i]=decodeMP3(&mp[i], inbuf[i], inbufsize[i],
			   decodebuf[i]+decoded[i], 32768, &add);
	  */
	  volume[i]=0;
	else
	  ret[i]=decodeMP3(&mp[i], NULL, 0,
			   decodebuf[i]+decoded[i], 32768, &add);
	
	if(ret[i]==MP3_OK)
	  decoded[i]+=add;
      }
    }
  }

  played+=outbufsize;
}

void closemp3() {
  int i=0;

  for(i=0; i<numfiles; i++) {
    /* Shutdown MP3 decoder */
    ExitMP3(&mp[i]);
    
    if(file[i]) {
      fclose(file[i]);
      file[i]=0;
    }
    if(inbuf[i]!=NULL) {
      free(inbuf[i]);
      inbuf[i]=NULL;
    }
  }
}
