/* tunnel.c
 *
 * Copyright 2001-2002 Vesa Halttunen (Vesuri/dA JoRMaS)
 *
 * This file is part of JRm-core.
 *
 * JRm-core is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JRm-core is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JRm-core; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include "effect.h"
#include "tunnel.h"

extern int sinus[1024];

void tunnel_run(void *d, long time) {
  tunnel_data *data=(tunnel_data *)d;
  int swidth, sheight, dwidth, dheight;
  unsigned int dx, dy, *source, *dest;
  double sx, sy, sa, sb, r;
  double sina, sinb, za, zb, zc, x, y, z;
  int tx, ty;

  /* Get local copies of the parameters */
  source=data->source->data;
  dest=data->dest->data;
  swidth=data->source->width;
  sheight=data->source->height;
  dwidth=data->dest->width;
  dheight=data->dest->height;
  sx=data->sx;
  sy=data->sy;
  sa=data->sa;
  sb=data->sb;
  r=data->r;

  zc=sx*sx+sy*sy-r*r;
  for(dy=0; dy<dheight; dy++) {
    sinb=sin(sb+(double)dy*M_PI/4/dheight-M_PI/8);
    for(dx=0; dx<dwidth; dx++) {
      sina=sin(sa+(double)dx*M_PI/4/dwidth-M_PI/8);
      za=sina*sina+sinb*sinb;
      zb=2*(sx*sina+sy*sinb);
      
      z=(-zb+sqrt(zb*zb-4*za*zc))/(2*za);
      x=sx+sina*z;
      y=sy+sinb*z;

      tx=((int)z+data->txadd)%swidth;
      ty=(int)(sheight*(atan(y/x)+M_PI/2)/M_PI);

      *dest++=source[ty*swidth+tx];
    }
  }
}

void tunnel_free(effect *effect) {
  free(effect->data);
}

void tunnel_new(effect *effect) {
  effect->run=tunnel_run;
  effect->new=tunnel_new;
  effect->freee=tunnel_free;
  effect->data=(tunnel_data *)calloc(sizeof(tunnel_data), 1);
}
