/* zoomrotate.c
 *
 * Copyright 2001-2002 Vesa Halttunen (Vesuri/dA JoRMaS)
 *
 * This file is part of JRm-core.
 *
 * JRm-core is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * JRm-core is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with JRm-core; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <math.h>
#include "effect.h"
#include "zoomrotate.h"

void zoomrotate_run(void *d, long time) {
  zoomrotate_data *data=(zoomrotate_data *)d;
  int angle, dx, dy, srcwidth, srcheight, destwidth, destheight;
  unsigned int px, py, zoom;
  unsigned int x, y, *source, *dest;

  /* Get local copies of the parameters */
  angle=data->angle, dx=data->dx, dy=data->dy;
  px=data->px, py=data->py, zoom=data->zoom;
  source=data->source->data;
  dest=data->dest->data;
  srcwidth=data->source->width;
  srcheight=data->source->height;
  destwidth=data->dest->width;
  destheight=data->dest->height;

  dx=zoom*cos((double)angle*2*M_PI/1024);
  dy=zoom*sin((double)angle*2*M_PI/1024);
  px=srcwidth/2*0x10000+destwidth/2*dx+destheight/2*dy;
  py=srcheight/2*0x10000-destwidth/2*dy+destheight/2*dx;
  px%=(srcwidth<<16)-1;
  py%=(srcheight<<16)-1;
 
  for(y=0; y<destheight; y++) {
    for(x=0; x<destwidth; x++) {
      *dest++=source[((py>>16)*srcwidth+(px>>16))];
      px-=dx;
      py+=dy;
      px%=(srcwidth<<16)-1;
      py%=(srcheight<<16)-1;
    }
    px-=(-destwidth*dx+dy);
    py+=(-destwidth*dy-dx);
    px%=(srcwidth<<16)-1;
    py%=(srcheight<<16)-1;
  }
}

void zoomrotate_free(effect *effect) {
  free(effect->data);
}

void zoomrotate_new(effect *effect) {
  effect->run=zoomrotate_run;
  effect->new=zoomrotate_new;
  effect->freee=zoomrotate_free;
  effect->data=(zoomrotate_data *)calloc(sizeof(zoomrotate_data), 1);
}
